// Implementation file for the IntAVLTree class
#include <iostream>
#include <cassert>
#include "IntAVLTree.h"
using namespace std;

//*************************************************************
// insert accepts a TreeNode pointer and a pointer to a node. *
// The function inserts the node into the tree pointed to by  *
// the TreeNode pointer. This function is called recursively. *
//*************************************************************

void IntAVLTree::insert(TreeNode *&nodePtr, TreeNode *&newNode, bool & taller)
{
	bool tallerLeftSubtree, tallerRightSubtree;

	taller = false;		// <<<<<<< Bug fix from original lab
	if (nodePtr == NULL) {
		nodePtr = newNode;				  // Insert the node.
		taller = true;
	} else if (newNode->value < nodePtr->value) { // insert into left subtree

		// ADD CODE HERE FOR PART A

	} else {// insert into right subtree
		insert(nodePtr->right, newNode, tallerRightSubtree);    
		if (tallerRightSubtree) {
			if (nodePtr->balance == TR) { 
				// Need rotation(s) to restore AVL height-balance property
				rebalanceRightSubtree(nodePtr);
				taller = false;
			} else if (nodePtr->balance == EQ) {
				nodePtr->balance = TR;
				taller = true;
			} else {  
				nodePtr->balance = EQ;
				taller = false;
			} // end if 

		} // end if (tallerRightSubtree)

	} // end if
}


// Performs a right (clockwise) rotation about the "oldSubtreeRoot"
void IntAVLTree::rotateRight(TreeNode *& oldSubtreeRoot)
{
	cout << "Rotating Right about: " << oldSubtreeRoot->value << endl;
	TreeNode * newSubtreeRoot = oldSubtreeRoot->left;
	oldSubtreeRoot->left = newSubtreeRoot->right;
	newSubtreeRoot->right = oldSubtreeRoot;
	oldSubtreeRoot = newSubtreeRoot;
} // end rotateRight

// Performs a left (counter-clockwise) rotation about the "oldSubtreeRoot"
void IntAVLTree::rotateLeft(TreeNode *& oldSubtreeRoot)
{
	cout << "Rotating Left about: " << oldSubtreeRoot->value << endl;

	// ADD CODE HERE FOR PART A

} // end rotateLeft

// Rebalances the right subtree after the longer right subtree of the pivot
// got even longer with the newly inserted node
void IntAVLTree::rebalanceRightSubtree(TreeNode *& pivot)
{
	TreeNode * pivotsRightChild = pivot->right;
	// Note:  The balance factor of all nodes below the pivot node have
	// already been adjusted to account for the new node!!!
	if (pivotsRightChild->balance == TR) { // RR case
		// Pre-set the balance factors to reflect the correct values after
		// the rotation in performed.
		pivot->balance = EQ;
		pivotsRightChild->balance = EQ;
		rotateLeft(pivot);
	} else if (pivotsRightChild->balance == EQ) { // an impossible case
		assert(pivotsRightChild->balance != EQ);
	} else { //pivotsRightChild->balance == TL  -- the RL cases
		TreeNode * leftChildOfPivotsRightChild = pivotsRightChild->left;
		// Pre-set the balance factors to reflect the correct values after
		// the rotations are performed.
		if (leftChildOfPivotsRightChild->balance == TR) { //RLR case
			leftChildOfPivotsRightChild->balance = EQ;
			pivotsRightChild->balance = EQ;
			pivot->balance = TL;
		} else if (leftChildOfPivotsRightChild->balance == EQ) { // RLNew case
			// the left child of the pivot's right child must be the new node
			pivotsRightChild->balance = EQ;
			pivot->balance = EQ;
		} else { // leftChildOfPivotsRightChild->balance == TL -- RLL case
			leftChildOfPivotsRightChild->balance = EQ;
			pivotsRightChild->balance = TR;
			pivot->balance = EQ;
		} // end if
		rotateRight(pivot->right); // Uses "pivot->right" instead of
		// "pivotsRightChild" to change the pivot node's right pointer
		rotateLeft(pivot);
	} // end if
} // end rebalanceRightSubtree


// Rebalances the left subtree after the longer left subtree of the pivot
// got even longer with the newly inserted node
void IntAVLTree::rebalanceLeftSubtree(TreeNode *& pivot)
{

	// ADD CODE HERE FOR PART A

} // end rebalanceLeftSubtree


//**********************************************************
// insertNode creates a new node to hold num as its value, *
// and passes it to the insert function.                   *
//**********************************************************

void IntAVLTree::insertNode(int num)
{
	TreeNode *newNode;      // Pointer to a new node.
	bool taller;		   // Indicates if the tree got taller by 1
	// upon insert of NewNode
	// Create a new node and store num in it.
	newNode = new TreeNode;
	newNode->value = num;
	newNode->left = newNode->right = NULL;
	newNode->balance = EQ;

	// Insert the node.
	insert(root, newNode, taller);
}

//***************************************************
// destroySubTree is called by the destructor. It   *
// deletes all nodes in the tree.                   *
//***************************************************

void IntAVLTree::destroySubTree(TreeNode *nodePtr)
{
	if (nodePtr)
	{
		if (nodePtr->left)
			destroySubTree(nodePtr->left);
		if (nodePtr->right)
			destroySubTree(nodePtr->right);
		delete nodePtr;
	}
}

//***************************************************
// searchNode determines if a value is present in   *
// the tree. If so, the function returns true.      *
// Otherwise, it returns false.                     *
//***************************************************

bool IntAVLTree::searchNode(int num)
{
	TreeNode *nodePtr = root;

	while (nodePtr)
	{
		if (nodePtr->value == num)
			return true;
		else if (num < nodePtr->value)
			nodePtr = nodePtr->left;
		else
			nodePtr = nodePtr->right;
	}
	return false;
}

//**********************************************
// remove calls deleteNode to delete the       *
// node whose value member is the same as num. *
//**********************************************

void IntAVLTree::remove(int num)
{
	deleteNode(num, root);
}


//********************************************
// deleteNode deletes the node whose value   *
// member is the same as num.                *
//********************************************

void IntAVLTree::deleteNode(int num, TreeNode *&nodePtr)
{
	if (num < nodePtr->value)
		deleteNode(num, nodePtr->left);
	else if (num > nodePtr->value)
		deleteNode(num, nodePtr->right);
	else
		makeDeletion(nodePtr);
}


//***********************************************************
// makeDeletion takes a reference to a pointer to the node  *
// that is to be deleted. The node is removed and the       *
// branches of the tree below the node are reattached.      *
//***********************************************************

void IntAVLTree::makeDeletion(TreeNode *&nodePtr)
{
	// Define a temporary pointer to use in reattaching
	// the left subtree.
	TreeNode *tempNodePtr;

	if (nodePtr == NULL)
		cout << "Cannot delete empty node.\n";
	else if (nodePtr->right == NULL)
	{
		tempNodePtr = nodePtr;
		nodePtr = nodePtr->left;   // Reattach the left child
		delete tempNodePtr;
	}
	else if (nodePtr->left == NULL)
	{
		tempNodePtr = nodePtr;
		nodePtr = nodePtr->right;  // Reattach the right child
		delete tempNodePtr;
	}
	// If the node has two children.
	else
	{
		// Move one node the right.
		tempNodePtr = nodePtr->right;
		// Go to the end left node.
		while (tempNodePtr->left)
			tempNodePtr = tempNodePtr->left;
		// Reattach the left subtree.
		tempNodePtr->left = nodePtr->left;
		tempNodePtr = nodePtr;
		// Reattach the right subtree.
		nodePtr = nodePtr->right;
		delete tempNodePtr;
	}
}

//****************************************************************
// The displayTree member function displays the values        *
// in the subtree pointed to by nodePtr, via inorder traversal.  *
//****************************************************************

void IntAVLTree::displayTree(TreeNode *nodePtr, int depth) const
{
	if (nodePtr)
	{
		displayTree(nodePtr->right, depth+1);
		if (nodePtr->right) {
			for (int count = 0; count < depth; count++) {
				cout << "   ";
			} // end for
			cout << " /" << endl;
		} // end if
		for (int count = 0; count < depth; count++) {
			cout << "   ";
		} 
		cout << nodePtr->value << endl;
		if (nodePtr->left) {
			for (int count = 0; count < depth; count++) {
				cout << "   ";
			} // end for
			cout << " \\" << endl;
		} // end if
		displayTree(nodePtr->left, depth+1);
	}
}
//****************************************************************
// The displayInOrder member function displays the values        *
// in the subtree pointed to by nodePtr, via inorder traversal.  *
//****************************************************************

void IntAVLTree::displayInOrder(TreeNode *nodePtr) const
{
	if (nodePtr)
	{
		displayInOrder(nodePtr->left);
		cout << nodePtr->value << endl;
		displayInOrder(nodePtr->right);
	}
}

//****************************************************************
// The displayPreOrder member function displays the values       *
// in the subtree pointed to by nodePtr, via preorder traversal. *
//****************************************************************

void IntAVLTree::displayPreOrder(TreeNode *nodePtr) const
{
	if (nodePtr)
	{
		cout << nodePtr->value << endl;
		displayPreOrder(nodePtr->left);     
		displayPreOrder(nodePtr->right);
	}
}

//****************************************************************
// The displayPostOrder member function displays the values      *
// in the subtree pointed to by nodePtr, via postorder traversal.*
//****************************************************************

void IntAVLTree::displayPostOrder(TreeNode *nodePtr) const
{
	if (nodePtr)
	{
		displayPostOrder(nodePtr->left);    
		displayPostOrder(nodePtr->right);
		cout << nodePtr->value << endl;
	}
}
