// IntAVLTree.h  Specification file for the IntAVLTree class
#ifndef INTAVLTREE_H
#define INTAVLTREE_H

class IntAVLTree
{
private:
	enum BalanceFactor { TL, EQ, TR };

   struct TreeNode
   {
      int value;         // The value in the node
	  BalanceFactor balance;  // The balance factor of the node
      TreeNode *left;    // Pointer to left child node
      TreeNode *right;   // Pointer to right child node
   };

   TreeNode *root;       // Pointer to the root node
   
   // Private member functions
   void insert(TreeNode *&, TreeNode *&, bool &);
   void rebalanceRightSubtree(TreeNode *&);
   void rebalanceLeftSubtree(TreeNode *&);
   void rotateRight(TreeNode *&);
   void rotateLeft(TreeNode *&);
   void destroySubTree(TreeNode *);
   void deleteNode(int, TreeNode *&);
   void makeDeletion(TreeNode *&);
   void displayTree(TreeNode *, int) const;
   void displayInOrder(TreeNode *) const;
   void displayPreOrder(TreeNode *) const;
   void displayPostOrder(TreeNode *) const;
   
public:
   // Constructor
   IntAVLTree()
      { root = NULL; }
      
   // Destructor
   ~IntAVLTree()
      { destroySubTree(root); }
      
   // Binary tree operations
   void insertNode(int);
   bool searchNode(int);
   void remove(int);

   
   void displayTree() const
      {  displayTree(root, 0); }
      
   void displayInOrder() const
      {  displayInOrder(root); }
      
   void displayPreOrder() const
      {  displayPreOrder(root); }
      
   void displayPostOrder() const
      {  displayPostOrder(root); }
};
#endif
